const fs = require('fs');

const express = require('express');

const app = express();
const server = require('http').createServer(app);
const io = require('socket.io')(server);
const path = require('path');

const config = require('./config');

const port = config.port || 3000;
// const filename = config.filename || 'knowledge.json';

// asking server to listen to 3000
server.listen(port, () => {
  console.log('Server listening at port %d', port);
});

// assign static path to public folder
app.use(express.static(path.join(__dirname, '/public/')));

const modRooms = [];

Object.keys(config.rooms).forEach((key) => {
  modRooms.push(`/${key}`);
});

app.get(modRooms, (req, res) => {
  res.sendFile(path.join(__dirname, '/public/mod.html'));
});

app.get(['/en', '/ml', '/hi'], (req, res) => {
  res.sendFile(path.join(__dirname, '/public/user.html'));
});

const userNamespace = io.of('/user');
const modNamespace = io.of(config.modNamespace);

const leafIds = [];
for (let i = 7; i < 332; i += 1) {
  const leafId = `path${i.toString()}`;
  i += 3;
  leafIds.push(leafId);
}

const roomLeafIds = {
  en: { leafIds: [...leafIds] },
  ml: { leafIds: [...leafIds] },
  hi: { leafIds: [...leafIds] },
};

const leavesData = {
  en: {},
  ml: {},
  hi: {},
  ma: {},
};

// object of structure {socketid1: {userName:"", leafText:"",leafId:""}, ...}

const fileSave = function fileSave(roomId) {
  const date = new Date();
  const filename = `sfcamp_${roomId}${date.getDate()}${date.getHours()}${date.getMinutes()}`;
  const filepath = path.join(__dirname, 'sfcamp', filename);
  if (Object.entries(leavesData[roomId]).length !== 0) {
    fs.writeFileSync(filepath, JSON.stringify(leavesData[roomId]), (err) => {
      if (err) return console.error(err);
    });
  }
};

const notifyEveryone = function notifyUserStateChangeToEveryone(roomId) {
  userNamespace.to(roomId).emit('leaves data', leavesData[roomId]);
  modNamespace.to(roomId).emit('leaves data', leavesData[roomId]);
  fileSave(roomId);
};

const updateLeavesData = function updateLeavesDataAndNotify(socketId, roomId, leafData) {
  if (!(leavesData[roomId][socketId].leafId)) {
    leavesData[roomId][socketId].leafId = roomLeafIds[roomId].leafIds.shift();
  }
  leavesData[roomId][socketId] = { ...leavesData[roomId][socketId], ...leafData };
  notifyEveryone(roomId);
};

const deleteLeaf = function deleteLeafAndNotifyEveryone(socketId, roomId) {
  const { leafId } = leavesData[roomId][socketId];
  delete leavesData[roomId][socketId].leafText;
  delete leavesData[roomId][socketId].leafId;
  roomLeafIds[roomId].leafIds.push(leafId);
  userNamespace.to(roomId).emit('leaf delete', leafId);
  modNamespace.to(roomId).emit('leaves data', leavesData[roomId]);
  fileSave(roomId);
};

userNamespace.on('connection', (socket) => {
  const socketId = socket.id;
  const { userName, roomId } = socket.handshake.query;
  if (roomId === 'en' || roomId === 'ml' || roomId === 'hi') {
    socket.join(roomId);

    leavesData[roomId][socketId] = { userName }; // user
    socket.emit('leaves data', leavesData[roomId]);
    modNamespace.to(roomId).emit('leaves data', leavesData[roomId]);

    socket.on('leaf text', (leafText) => {
      const leafData = {};
      leafData.leafText = leafText;
      if (!(leavesData[roomId][socketId].leafId)) {
        leafData.leafId = roomLeafIds[roomId].leafIds.shift();
      }

      updateLeavesData(socketId, roomId, leafData);
    });
  } else {
    socket.disconnect();
  }
});

modNamespace.on('connection', (socket) => {
  const { roomName } = socket.handshake.query;
  const roomId = config.rooms[roomName];
  if (roomId) {
    socket.join(roomId);

    modNamespace.to(roomId).emit('leaves data', leavesData[roomId]);
    socket.on('allow user', (socketId) => {
      userNamespace.to(socketId).emit('enter knowledge');
    });
    socket.on('leaf delete', (socketId) => {
      deleteLeaf(socketId, roomId);
    });
    socket.on('leaf update', (data) => {
      updateLeavesData(data.socketId, roomId, data.leafData);
    });
    socket.on('user delete', (socketId) => {
      deleteLeaf(socketId, roomId);
      userNamespace.sockets.get(socketId).disconnect();
      delete leavesData[roomId][socketId];
      modNamespace.to(roomId).emit('leaves data', leavesData[roomId]);
      fileSave(roomId);
    });
    socket.on('save exit', () => { // save data into file
      fileSave(roomId);
      userNamespace.in(roomId).disconnectSockets();
      Object.keys(leavesData[roomId]).forEach((key) => {
        delete leavesData[roomId][key];
      });
      roomLeafIds[roomId].leafIds = leafIds;
      notifyEveryone(roomId);
    });
  } else {
    socket.disconnect();
  }
});
